@@ -6,8 +6,10 @@
 #define V8_EXECUTION_POINTER_AUTHENTICATION_DUMMY_H_
 
 #include "include/v8-internal.h"
+#include "src/base/logging.h"
 #include "src/base/macros.h"
 #include "src/execution/pointer-authentication.h"
+#include "src/flags/flags.h"
 
 namespace v8 {
 namespace internal {
@@ -36,6 +38,17 @@ V8_INLINE Address PointerAuthentication::SignAndCheckPC(Isolate*, Address pc,
   return pc;
 }
 
+V8_INLINE Address PointerAuthentication::MoveSignedPC(Isolate*, Address pc,
+                                                      Address, Address) {
+#if V8_ENABLE_WEBASSEMBLY
+  // Only used by wasm deoptimizations and growable stacks.
+  CHECK(v8_flags.wasm_deopt || v8_flags.experimental_wasm_growable_stacks);
+  return pc;
+#else
+  UNREACHABLE();
+#endif
+}
+
 }  // namespace internal
 }  // namespace v8
 
