@@ -2,6 +2,9 @@
 // Use of this source code is governed by a BSD-style license that can be
 // found in the LICENSE file.
 
+#ifndef V8_EXECUTION_S390_SIMULATOR_S390_H_
+#define V8_EXECUTION_S390_SIMULATOR_S390_H_
+
 // Declares a Simulator for S390 instructions if we are not generating a native
 // S390 binary. This Simulator allows us to run and debug S390 code generation
 // on regular desktop machines.
@@ -9,9 +12,6 @@
 // which will start execution in the Simulator or forwards to the real entry
 // on a S390 hardware platform.
 
-#ifndef V8_EXECUTION_S390_SIMULATOR_S390_H_
-#define V8_EXECUTION_S390_SIMULATOR_S390_H_
-
 // globals.h defines USE_SIMULATOR.
 #include "src/common/globals.h"
 
@@ -24,6 +24,10 @@
 #include "src/execution/simulator-base.h"
 #include "src/utils/allocation.h"
 
+namespace heap::base {
+class StackVisitor;
+}
+
 namespace v8 {
 namespace internal {
 
@@ -163,9 +167,15 @@ class Simulator : public SimulatorBase {
   // Accessor to the internal simulator stack area. Adds a safety
   // margin to prevent overflows.
   uintptr_t StackLimit(uintptr_t c_limit) const;
-  // Return current stack view, without additional safety margins.
+
+  uintptr_t StackBase() const;
+
+  // Return central stack view, without additional safety margins.
   // Users, for example wasm::StackMemory, can add their own.
-  base::Vector<uint8_t> GetCurrentStackView() const;
+  base::Vector<uint8_t> GetCentralStackView() const;
+  static constexpr int JSStackLimitMargin() { return kStackProtectionSize; }
+
+  void IterateRegistersAndStack(::heap::base::StackVisitor* visitor);
 
   // Executes S390 instructions until the PC reaches end_sim_pc.
   void Execute();
@@ -181,10 +191,10 @@ class Simulator : public SimulatorBase {
   double CallFPReturnsDouble(Address entry, double d0, double d1);
 
   // Push an address onto the JS stack.
-  uintptr_t PushAddress(uintptr_t address);
+  V8_EXPORT_PRIVATE uintptr_t PushAddress(uintptr_t address);
 
   // Pop an address from the JS stack.
-  uintptr_t PopAddress();
+  V8_EXPORT_PRIVATE uintptr_t PopAddress();
 
   // Debugger input.
   void set_last_debugger_input(char* input);
@@ -202,6 +212,11 @@ class Simulator : public SimulatorBase {
   // below (bad_lr, end_sim_pc).
   bool has_bad_pc() const;
 
+  // Manage instruction tracing.
+  bool InstructionTracingEnabled();
+
+  void ToggleInstructionTracing();
+
   enum special_values {
     // Known bad pc value to ensure that the simulator does not execute
     // without being properly setup.
@@ -235,6 +250,10 @@ class Simulator : public SimulatorBase {
   void SoftwareInterrupt(Instruction* instr);
   void DebugAtNextPC();
 
+  // Take a copy of v8 simulator tracing flag because flags are frozen after
+  // start.
+  bool instruction_tracing_ = v8_flags.trace_sim;
+
   // Stop helper functions.
   inline bool isStopInstruction(Instruction* instr);
   inline bool isWatchedStop(uint32_t bkpt_code);
@@ -421,16 +440,12 @@ class Simulator : public SimulatorBase {
   // Special register to track PC.
   intptr_t special_reg_pc_;
 
-  // Simulator support.
+  // Simulator support for the stack.
   uint8_t* stack_;
   static const size_t kStackProtectionSize = 256 * kSystemPointerSize;
   // This includes a protection margin at each end of the stack area.
   static size_t AllocatedStackSize() {
-#if V8_TARGET_ARCH_S390X
     size_t stack_size = v8_flags.sim_stack_size * KB;
-#else
-    size_t stack_size = MB;  // allocate 1MB for stack
-#endif
     return stack_size + (2 * kStackProtectionSize);
   }
   static size_t UsableStackSize() {
