@@ -267,8 +267,9 @@ void V8ProfilerAgentImpl::restore() {
       bool updatesAllowed = m_state->booleanProperty(
           ProfilerAgentState::preciseCoverageAllowTriggeredUpdates, false);
       double timestamp;
-      startPreciseCoverage(Maybe<bool>(callCount), Maybe<bool>(detailed),
-                           Maybe<bool>(updatesAllowed), &timestamp);
+      startPreciseCoverage(std::optional<bool>(callCount),
+                           std::optional<bool>(detailed),
+                           std::optional<bool>(updatesAllowed), &timestamp);
     }
   }
 }
@@ -301,8 +302,8 @@ Response V8ProfilerAgentImpl::stop(
 }
 
 Response V8ProfilerAgentImpl::startPreciseCoverage(
-    Maybe<bool> callCount, Maybe<bool> detailed,
-    Maybe<bool> allowTriggeredUpdates, double* out_timestamp) {
+    std::optional<bool> callCount, std::optional<bool> detailed,
+    std::optional<bool> allowTriggeredUpdates, double* out_timestamp) {
   if (!m_enabled) return Response::ServerError("Profiler is not enabled");
   *out_timestamp = v8::base::TimeTicks::Now().since_origin().InSecondsF();
   bool callCountValue = callCount.value_or(false);
