@@ -24,10 +24,10 @@ bool CanonicalNumericIndexString(Isolate* isolate,
   *is_minus_zero = false;
   if (lookup_key.is_element()) return true;
 
-  Handle<String> key = Handle<String>::cast(lookup_key.name());
+  DirectHandle<String> key = Cast<String>(lookup_key.name());
 
   // 3. Let n be ! ToNumber(argument).
-  Handle<Object> result = String::ToNumber(isolate, key);
+  DirectHandle<Object> result = String::ToNumber(isolate, key);
   if (IsMinusZero(*result)) {
     // 2. If argument is "-0", return -0𝔽.
     // We are not performing SaveValue check for -0 because it'll be rejected
@@ -35,7 +35,8 @@ bool CanonicalNumericIndexString(Isolate* isolate,
     *is_minus_zero = true;
   } else {
     // 4. If SameValue(! ToString(n), argument) is false, return undefined.
-    Handle<String> str = Object::ToString(isolate, result).ToHandleChecked();
+    DirectHandle<String> str =
+        Object::ToString(isolate, result).ToHandleChecked();
     // Avoid treating strings like "2E1" and "20" as the same key.
     if (!Object::SameValue(*str, *key)) return false;
   }
@@ -46,51 +47,54 @@ bool CanonicalNumericIndexString(Isolate* isolate,
 void JSArrayBuffer::Setup(SharedFlag shared, ResizableFlag resizable,
                           std::shared_ptr<BackingStore> backing_store,
                           Isolate* isolate) {
+  auto finish_setup = [shared, isolate]() {
+    // Count usage may lead to a blink allocation, through the callback, which
+    // may trigger a GC. It is important to delay this, until the array buffer
+    // is properly initialized.
+    if (shared == SharedFlag::kShared) {
+      isolate->CountUsage(
+          v8::Isolate::UseCounterFeature::kSharedArrayBufferConstructed);
+    }
+  };
   clear_padding();
+  init_extension();
   set_detach_key(ReadOnlyRoots(isolate).undefined_value());
   set_bit_field(0);
   set_is_shared(shared == SharedFlag::kShared);
   set_is_resizable_by_js(resizable == ResizableFlag::kResizable);
   set_is_detachable(shared != SharedFlag::kShared);
+  SetupLazilyInitializedCppHeapPointerField(
+      JSAPIObjectWithEmbedderSlots::kCppHeapWrappableOffset);
   for (int i = 0; i < v8::ArrayBuffer::kEmbedderFieldCount; i++) {
     SetEmbedderField(i, Smi::zero());
   }
-  set_extension(nullptr);
   if (!backing_store) {
     set_backing_store(isolate, EmptyBackingStoreBuffer());
     set_byte_length(0);
     set_max_byte_length(0);
-  } else {
-    Attach(std::move(backing_store));
+    finish_setup();
+    return;
   }
-  if (shared == SharedFlag::kShared) {
-    isolate->CountUsage(
-        v8::Isolate::UseCounterFeature::kSharedArrayBufferConstructed);
-  }
-}
-
-void JSArrayBuffer::Attach(std::shared_ptr<BackingStore> backing_store) {
-  DCHECK_NOT_NULL(backing_store);
+  // Rest of the code here deals with attaching the BackingStore.
   DCHECK_EQ(is_shared(), backing_store->is_shared());
-  DCHECK_EQ(is_resizable_by_js(), backing_store->is_resizable_by_js());
+  DCHECK((is_resizable_by_js() == backing_store->is_resizable_by_js()) ||
+         (backing_store->is_wasm_memory() && is_shared()));
   DCHECK_IMPLIES(
       !backing_store->is_wasm_memory() && !backing_store->is_resizable_by_js(),
       backing_store->byte_length() == backing_store->max_byte_length());
-  DCHECK(!was_detached());
-  Isolate* isolate = GetIsolate();
 
   void* backing_store_buffer = backing_store->buffer_start();
   // Wasm memory always needs a backing store; this is guaranteed by reserving
   // at least one page for the BackingStore (so {IsEmpty()} is always false).
-  CHECK_IMPLIES(backing_store->is_wasm_memory(), !backing_store->IsEmpty());
+  DCHECK_IMPLIES(backing_store->is_wasm_memory(), !backing_store->IsEmpty());
   // Non-empty backing stores must start at a non-null pointer.
-  DCHECK_IMPLIES(backing_store_buffer == nullptr, backing_store->IsEmpty());
-  // Empty backing stores can be backed by a null pointer or an externally
-  // provided pointer: Either is acceptable. If pointers are sandboxed then
-  // null pointers must be replaced by a special null entry.
-  if (V8_ENABLE_SANDBOX_BOOL && !backing_store_buffer) {
-    backing_store_buffer = EmptyBackingStoreBuffer();
-  }
+  DCHECK_IMPLIES(backing_store_buffer == EmptyBackingStoreBuffer(),
+                 backing_store->IsEmpty());
+  // Empty backing stores can be backed by an empty buffer pointer or by an
+  // externally provided pointer: Either is acceptable. However, the pointer
+  // must always point into the sandbox, so nullptr is not acceptable if the
+  // sandbox is enabled.
+  DCHECK_IMPLIES(V8_ENABLE_SANDBOX_BOOL, backing_store_buffer != nullptr);
   set_backing_store(isolate, backing_store_buffer);
 
   // GSABs need to read their byte_length from the BackingStore. Maintain the
@@ -99,6 +103,7 @@ void JSArrayBuffer::Attach(std::shared_ptr<BackingStore> backing_store) {
       (is_shared() && is_resizable_by_js()) ? 0 : backing_store->byte_length();
   CHECK_LE(backing_store->byte_length(), kMaxByteLength);
   set_byte_length(byte_len);
+
   // For Wasm memories, it is possible for the backing store maximum to be
   // different from the JSArrayBuffer maximum. The maximum pages allowed on a
   // Wasm memory are tracked on the Wasm memory object, and not the
@@ -106,20 +111,21 @@ void JSArrayBuffer::Attach(std::shared_ptr<BackingStore> backing_store) {
   auto max_byte_len = is_resizable_by_js() ? backing_store->max_byte_length()
                                            : backing_store->byte_length();
   set_max_byte_length(max_byte_len);
-  if (backing_store->is_wasm_memory()) set_is_detachable(false);
-  ArrayBufferExtension* extension = EnsureExtension();
-  size_t bytes = backing_store->PerIsolateAccountingLength();
-  extension->set_accounting_length(bytes);
-  extension->set_backing_store(std::move(backing_store));
-  isolate->heap()->AppendArrayBufferExtension(*this, extension);
+
+  if (backing_store->is_wasm_memory()) {
+    set_is_detachable(false);
+  }
+
+  CreateExtension(isolate, std::move(backing_store));
+  finish_setup();
 }
 
-Maybe<bool> JSArrayBuffer::Detach(Handle<JSArrayBuffer> buffer,
+Maybe<bool> JSArrayBuffer::Detach(DirectHandle<JSArrayBuffer> buffer,
                                   bool force_for_wasm_memory,
-                                  Handle<Object> maybe_key) {
+                                  DirectHandle<Object> maybe_key) {
   Isolate* const isolate = buffer->GetIsolate();
 
-  Handle<Object> detach_key = handle(buffer->detach_key(), isolate);
+  DirectHandle<Object> detach_key(buffer->detach_key(), isolate);
 
   bool key_mismatch = false;
 
@@ -158,7 +164,7 @@ void JSArrayBuffer::DetachInternal(bool force_for_wasm_memory,
 
   if (extension) {
     DisallowGarbageCollection disallow_gc;
-    isolate->heap()->DetachArrayBufferExtension(*this, extension);
+    isolate->heap()->DetachArrayBufferExtension(extension);
     std::shared_ptr<BackingStore> backing_store = RemoveExtension();
     CHECK_IMPLIES(force_for_wasm_memory, backing_store->is_wasm_memory());
   }
@@ -180,7 +186,7 @@ size_t JSArrayBuffer::GsabByteLength(Isolate* isolate,
   DisallowGarbageCollection no_gc;
   DisallowJavascriptExecution no_js(isolate);
   Tagged<JSArrayBuffer> buffer =
-      JSArrayBuffer::cast(Tagged<Object>(raw_array_buffer));
+      Cast<JSArrayBuffer>(Tagged<Object>(raw_array_buffer));
   CHECK(buffer->is_resizable_by_js());
   CHECK(buffer->is_shared());
   return buffer->GetBackingStore()->byte_length(std::memory_order_seq_cst);
@@ -216,12 +222,43 @@ Maybe<bool> JSArrayBuffer::GetResizableBackingStorePageConfiguration(
   return Just(true);
 }
 
-ArrayBufferExtension* JSArrayBuffer::EnsureExtension() {
-  ArrayBufferExtension* extension = this->extension();
-  if (extension != nullptr) return extension;
+// static
+std::optional<MessageTemplate>
+JSArrayBuffer::GetResizableBackingStorePageConfigurationImpl(
+    Isolate* isolate, size_t byte_length, size_t max_byte_length,
+    size_t* page_size, size_t* initial_pages, size_t* max_pages) {
+  DCHECK_NOT_NULL(page_size);
+  DCHECK_NOT_NULL(initial_pages);
+  DCHECK_NOT_NULL(max_pages);
 
-  extension = new ArrayBufferExtension(std::shared_ptr<BackingStore>());
+  *page_size = AllocatePageSize();
+
+  if (!RoundUpToPageSize(byte_length, *page_size, JSArrayBuffer::kMaxByteLength,
+                         initial_pages)) {
+    return MessageTemplate::kInvalidArrayBufferLength;
+  }
+
+  if (!RoundUpToPageSize(max_byte_length, *page_size,
+                         JSArrayBuffer::kMaxByteLength, max_pages)) {
+    return MessageTemplate::kInvalidArrayBufferMaxLength;
+  }
+  return {};
+}
+
+ArrayBufferExtension* JSArrayBuffer::CreateExtension(
+    Isolate* isolate, std::shared_ptr<BackingStore> backing_store) {
+  // `Heap::InYoungGeneration` during full GC with sticky markbits is generally
+  // inaccurate. However, a full GC will sweep both lists and promote all to
+  // old, so it doesn't matter which list initially holds the extension in this
+  // case.
+  const auto age =
+      HeapLayout::InYoungGeneration(UncheckedCast<JSArrayBuffer>(*this))
+          ? ArrayBufferExtension::Age::kYoung
+          : ArrayBufferExtension::Age::kOld;
+  ArrayBufferExtension* extension =
+      new ArrayBufferExtension(std::move(backing_store), age);
   set_extension(extension);
+  isolate->heap()->AppendArrayBufferExtension(extension);
   return extension;
 }
 
@@ -245,6 +282,7 @@ void JSArrayBuffer::MarkExtension() {
 void JSArrayBuffer::YoungMarkExtension() {
   ArrayBufferExtension* extension = this->extension();
   if (extension) {
+    DCHECK_EQ(ArrayBufferExtension::Age::kYoung, extension->age());
     extension->YoungMark();
   }
 }
@@ -258,9 +296,9 @@ void JSArrayBuffer::YoungMarkExtensionPromoted() {
 
 Handle<JSArrayBuffer> JSTypedArray::GetBuffer() {
   Isolate* isolate = GetIsolate();
-  Handle<JSTypedArray> self(*this, isolate);
+  DirectHandle<JSTypedArray> self(*this, isolate);
   DCHECK(IsTypedArrayOrRabGsabTypedArrayElementsKind(self->GetElementsKind()));
-  Handle<JSArrayBuffer> array_buffer(JSArrayBuffer::cast(self->buffer()),
+  Handle<JSArrayBuffer> array_buffer(Cast<JSArrayBuffer>(self->buffer()),
                                      isolate);
   if (!is_on_heap()) {
     // Already is off heap, so return the existing buffer.
@@ -301,8 +339,8 @@ Handle<JSArrayBuffer> JSTypedArray::GetBuffer() {
 // ES#sec-integer-indexed-exotic-objects-defineownproperty-p-desc
 // static
 Maybe<bool> JSTypedArray::DefineOwnProperty(Isolate* isolate,
-                                            Handle<JSTypedArray> o,
-                                            Handle<Object> key,
+                                            DirectHandle<JSTypedArray> o,
+                                            DirectHandle<Object> key,
                                             PropertyDescriptor* desc,
                                             Maybe<ShouldThrow> should_throw) {
   DCHECK(IsName(*key) || IsNumber(*key));
@@ -355,7 +393,7 @@ Maybe<bool> JSTypedArray::DefineOwnProperty(Isolate* isolate,
         if (!desc->has_configurable()) desc->set_configurable(true);
         if (!desc->has_enumerable()) desc->set_enumerable(true);
         if (!desc->has_writable()) desc->set_writable(true);
-        Handle<Object> value = desc->value();
+        DirectHandle<Object> value = desc->value();
         LookupIterator it(isolate, o, index, LookupIterator::OWN);
         RETURN_ON_EXCEPTION_VALUE(
             isolate,
@@ -406,7 +444,7 @@ size_t JSTypedArray::LengthTrackingGsabBackedTypedArrayLength(
   // in bounds checks to minimize the need for calling this function.
   DisallowGarbageCollection no_gc;
   DisallowJavascriptExecution no_js(isolate);
-  Tagged<JSTypedArray> array = JSTypedArray::cast(Tagged<Object>(raw_array));
+  Tagged<JSTypedArray> array = Cast<JSTypedArray>(Tagged<Object>(raw_array));
   CHECK(array->is_length_tracking());
   Tagged<JSArrayBuffer> buffer = array->buffer();
   CHECK(buffer->is_resizable_by_js());
@@ -418,35 +456,38 @@ size_t JSTypedArray::LengthTrackingGsabBackedTypedArrayLength(
   return (backing_byte_length - array->byte_offset()) / element_byte_size;
 }
 
-size_t JSTypedArray::GetVariableLengthOrOutOfBounds(bool& out_of_bounds) const {
+size_t JSTypedArray::GetVariableByteLengthOrOutOfBounds(
+    bool& out_of_bounds) const {
   DCHECK(!WasDetached());
+  size_t own_byte_offset = byte_offset();
   if (is_length_tracking()) {
     if (is_backed_by_rab()) {
-      if (byte_offset() > buffer()->byte_length()) {
+      size_t buffer_byte_length = buffer()->byte_length();
+      if (own_byte_offset > buffer_byte_length) {
         out_of_bounds = true;
         return 0;
       }
-      return (buffer()->byte_length() - byte_offset()) / element_size();
+      return (buffer_byte_length - own_byte_offset);
     }
-    if (byte_offset() >
-        buffer()->GetBackingStore()->byte_length(std::memory_order_seq_cst)) {
-      out_of_bounds = true;
-      return 0;
-    }
-    return (buffer()->GetBackingStore()->byte_length(
-                std::memory_order_seq_cst) -
-            byte_offset()) /
-           element_size();
+    // GSAB-backed TypedArrays can't be out of bounds.
+    size_t buffer_byte_length =
+        buffer()->GetBackingStore()->byte_length(std::memory_order_seq_cst);
+    SBXCHECK(own_byte_offset <= buffer_byte_length);
+    return buffer_byte_length - own_byte_offset;
   }
   DCHECK(is_backed_by_rab());
-  size_t array_length = LengthUnchecked();
-  // The sum can't overflow, since we have managed to allocate the
-  // JSTypedArray.
-  if (byte_offset() + array_length * element_size() > buffer()->byte_length()) {
+  size_t own_byte_length = byte_length();
+  size_t buffer_byte_length = buffer()->byte_length();
+  if (own_byte_length > buffer_byte_length ||
+      own_byte_offset > buffer_byte_length - own_byte_length) {
     out_of_bounds = true;
     return 0;
   }
-  return array_length;
+  return own_byte_length;
+}
+
+size_t JSTypedArray::GetVariableLengthOrOutOfBounds(bool& out_of_bounds) const {
+  return GetVariableByteLengthOrOutOfBounds(out_of_bounds) / element_size();
 }
 
 }  // namespace internal
