@@ -77,6 +77,8 @@ UDisplayContext ToUDisplayContext(JSDisplayNames::Style style) {
 // Abstract class for all different types.
 class DisplayNamesInternal {
  public:
+  static constexpr ExternalPointerTag kManagedTag = kDisplayNamesInternalTag;
+
   DisplayNamesInternal() = default;
   virtual ~DisplayNamesInternal() = default;
   virtual const char* type() const = 0;
@@ -230,7 +232,7 @@ class KeyValueDisplayNames : public LocaleDisplayNamesCommon {
     locale_display_names()->keyValueDisplayName(key_.c_str(), code_str.c_str(),
                                                 result);
     // Work around the issue that the keyValueDisplayNames ignore no
-    // substituion and always fallback.
+    // substitution and always fallback.
     if (prevent_fallback_ && (result.length() == 3) &&
         (code_str.length() == 3) &&
         (result == icu::UnicodeString(code_str.c_str(), -1, US_INV))) {
@@ -399,25 +401,23 @@ DisplayNamesInternal* CreateInternal(const icu::Locale& locale,
 }  // anonymous namespace
 
 // ecma402 #sec-Intl.DisplayNames
-MaybeHandle<JSDisplayNames> JSDisplayNames::New(Isolate* isolate,
-                                                Handle<Map> map,
-                                                Handle<Object> locales,
-                                                Handle<Object> input_options) {
+MaybeDirectHandle<JSDisplayNames> JSDisplayNames::New(
+    Isolate* isolate, DirectHandle<Map> map, DirectHandle<Object> locales,
+    DirectHandle<Object> input_options) {
   const char* service = "Intl.DisplayNames";
   Factory* factory = isolate->factory();
 
-  Handle<JSReceiver> options;
+  DirectHandle<JSReceiver> options;
   // 3. Let requestedLocales be ? CanonicalizeLocaleList(locales).
   Maybe<std::vector<std::string>> maybe_requested_locales =
       Intl::CanonicalizeLocaleList(isolate, locales);
-  MAYBE_RETURN(maybe_requested_locales, Handle<JSDisplayNames>());
+  MAYBE_RETURN(maybe_requested_locales, DirectHandle<JSDisplayNames>());
   std::vector<std::string> requested_locales =
       maybe_requested_locales.FromJust();
 
   // 4. Let options be ? GetOptionsObject(options).
   ASSIGN_RETURN_ON_EXCEPTION(isolate, options,
-                             GetOptionsObject(isolate, input_options, service),
-                             JSDisplayNames);
+                             GetOptionsObject(isolate, input_options, service));
 
   // Note: No need to create a record. It's not observable.
   // 5. Let opt be a new Record.
@@ -428,7 +428,7 @@ MaybeHandle<JSDisplayNames> JSDisplayNames::New(Isolate* isolate,
   // "lookup", "best fit" », "best fit").
   Maybe<Intl::MatcherOption> maybe_locale_matcher =
       Intl::GetLocaleMatcher(isolate, options, service);
-  MAYBE_RETURN(maybe_locale_matcher, MaybeHandle<JSDisplayNames>());
+  MAYBE_RETURN(maybe_locale_matcher, MaybeDirectHandle<JSDisplayNames>());
 
   // 8. Set opt.[[localeMatcher]] to matcher.
   Intl::MatcherOption matcher = maybe_locale_matcher.FromJust();
@@ -443,8 +443,7 @@ MaybeHandle<JSDisplayNames> JSDisplayNames::New(Isolate* isolate,
       Intl::ResolveLocale(isolate, JSDisplayNames::GetAvailableLocales(),
                           requested_locales, matcher, relevant_extension_keys);
   if (maybe_resolve_locale.IsNothing()) {
-    THROW_NEW_ERROR(isolate, NewRangeError(MessageTemplate::kIcuError),
-                    JSDisplayNames);
+    THROW_NEW_ERROR(isolate, NewRangeError(MessageTemplate::kIcuError));
   }
   Intl::ResolvedLocale r = maybe_resolve_locale.FromJust();
 
@@ -455,7 +454,7 @@ MaybeHandle<JSDisplayNames> JSDisplayNames::New(Isolate* isolate,
   Maybe<Style> maybe_style = GetStringOption<Style>(
       isolate, options, "style", service, {"long", "short", "narrow"},
       {Style::kLong, Style::kShort, Style::kNarrow}, Style::kLong);
-  MAYBE_RETURN(maybe_style, MaybeHandle<JSDisplayNames>());
+  MAYBE_RETURN(maybe_style, MaybeDirectHandle<JSDisplayNames>());
   Style style_enum = maybe_style.FromJust();
 
   // 11. Set displayNames.[[Style]] to style.
@@ -469,13 +468,12 @@ MaybeHandle<JSDisplayNames> JSDisplayNames::New(Isolate* isolate,
       {Type::kLanguage, Type::kRegion, Type::kScript, Type::kCurrency,
        Type::kCalendar, Type::kDateTimeField},
       Type::kUndefined);
-  MAYBE_RETURN(maybe_type, MaybeHandle<JSDisplayNames>());
+  MAYBE_RETURN(maybe_type, MaybeDirectHandle<JSDisplayNames>());
   Type type_enum = maybe_type.FromJust();
 
   // 13. If type is undefined, throw a TypeError exception.
   if (type_enum == Type::kUndefined) {
-    THROW_NEW_ERROR(isolate, NewTypeError(MessageTemplate::kInvalidArgument),
-                    JSDisplayNames);
+    THROW_NEW_ERROR(isolate, NewTypeError(MessageTemplate::kInvalidArgument));
   }
 
   // 14. Set displayNames.[[Type]] to type.
@@ -485,7 +483,7 @@ MaybeHandle<JSDisplayNames> JSDisplayNames::New(Isolate* isolate,
   Maybe<Fallback> maybe_fallback = GetStringOption<Fallback>(
       isolate, options, "fallback", service, {"code", "none"},
       {Fallback::kCode, Fallback::kNone}, Fallback::kCode);
-  MAYBE_RETURN(maybe_fallback, MaybeHandle<JSDisplayNames>());
+  MAYBE_RETURN(maybe_fallback, MaybeDirectHandle<JSDisplayNames>());
   Fallback fallback_enum = maybe_fallback.FromJust();
 
   // 16. Set displayNames.[[Fallback]] to fallback.
@@ -498,7 +496,7 @@ MaybeHandle<JSDisplayNames> JSDisplayNames::New(Isolate* isolate,
           isolate, options, "languageDisplay", service, {"dialect", "standard"},
           {LanguageDisplay::kDialect, LanguageDisplay::kStandard},
           LanguageDisplay::kDialect);
-  MAYBE_RETURN(maybe_language_display, MaybeHandle<JSDisplayNames>());
+  MAYBE_RETURN(maybe_language_display, MaybeDirectHandle<JSDisplayNames>());
   // 25. If type is "language", then
   if (type_enum == Type::kLanguage) {
     // a. Set displayNames.[[LanguageDisplay]] to languageDisplay.
@@ -527,19 +525,18 @@ MaybeHandle<JSDisplayNames> JSDisplayNames::New(Isolate* isolate,
 
   // Set displayNames.[[Fields]] to styleFields.
 
-  DisplayNamesInternal* internal = CreateInternal(
+  std::shared_ptr<DisplayNamesInternal> internal{CreateInternal(
       icu_locale, style_enum, type_enum, fallback_enum == Fallback::kCode,
-      language_display_enum == LanguageDisplay::kDialect);
+      language_display_enum == LanguageDisplay::kDialect)};
   if (internal == nullptr) {
-    THROW_NEW_ERROR(isolate, NewTypeError(MessageTemplate::kIcuError),
-                    JSDisplayNames);
+    THROW_NEW_ERROR(isolate, NewTypeError(MessageTemplate::kIcuError));
   }
 
-  Handle<Managed<DisplayNamesInternal>> managed_internal =
-      Managed<DisplayNamesInternal>::FromRawPtr(isolate, 0, internal);
+  DirectHandle<Managed<DisplayNamesInternal>> managed_internal =
+      Managed<DisplayNamesInternal>::From(isolate, 0, std::move(internal));
 
-  Handle<JSDisplayNames> display_names =
-      Handle<JSDisplayNames>::cast(factory->NewFastOrSlowJSObjectFromMap(map));
+  DirectHandle<JSDisplayNames> display_names =
+      Cast<JSDisplayNames>(factory->NewFastOrSlowJSObjectFromMap(map));
   display_names->set_flags(0);
   display_names->set_style(style_enum);
   display_names->set_fallback(fallback_enum);
@@ -553,22 +550,25 @@ MaybeHandle<JSDisplayNames> JSDisplayNames::New(Isolate* isolate,
 }
 
 // ecma402 #sec-Intl.DisplayNames.prototype.resolvedOptions
-Handle<JSObject> JSDisplayNames::ResolvedOptions(
-    Isolate* isolate, Handle<JSDisplayNames> display_names) {
+DirectHandle<JSObject> JSDisplayNames::ResolvedOptions(
+    Isolate* isolate, DirectHandle<JSDisplayNames> display_names) {
   Factory* factory = isolate->factory();
   // 4. Let options be ! ObjectCreate(%ObjectPrototype%).
-  Handle<JSObject> options = factory->NewJSObject(isolate->object_function());
+  DirectHandle<JSObject> options =
+      factory->NewJSObject(isolate->object_function());
 
   DisplayNamesInternal* internal = display_names->internal()->raw();
 
   Maybe<std::string> maybe_locale = Intl::ToLanguageTag(internal->locale());
   DCHECK(maybe_locale.IsJust());
-  Handle<String> locale = isolate->factory()->NewStringFromAsciiChecked(
+  DirectHandle<String> locale = isolate->factory()->NewStringFromAsciiChecked(
       maybe_locale.FromJust().c_str());
-  Handle<String> style = display_names->StyleAsString();
-  Handle<String> type = factory->NewStringFromAsciiChecked(internal->type());
-  Handle<String> fallback = display_names->FallbackAsString();
-  Handle<String> language_display = display_names->LanguageDisplayAsString();
+  DirectHandle<String> style = display_names->StyleAsString(isolate);
+  DirectHandle<String> type =
+      factory->NewStringFromAsciiChecked(internal->type());
+  DirectHandle<String> fallback = display_names->FallbackAsString(isolate);
+  DirectHandle<String> language_display =
+      display_names->LanguageDisplayAsString(isolate);
 
   Maybe<bool> maybe_create_locale = JSReceiver::CreateDataProperty(
       isolate, options, factory->locale_string(), locale, Just(kDontThrow));
@@ -603,16 +603,16 @@ Handle<JSObject> JSDisplayNames::ResolvedOptions(
 }
 
 // ecma402 #sec-Intl.DisplayNames.prototype.of
-MaybeHandle<Object> JSDisplayNames::Of(Isolate* isolate,
-                                       Handle<JSDisplayNames> display_names,
-                                       Handle<Object> code_obj) {
-  Handle<String> code;
-  ASSIGN_RETURN_ON_EXCEPTION(isolate, code, Object::ToString(isolate, code_obj),
-                             Object);
+MaybeDirectHandle<Object> JSDisplayNames::Of(
+    Isolate* isolate, DirectHandle<JSDisplayNames> display_names,
+    Handle<Object> code_obj) {
+  DirectHandle<String> code;
+  ASSIGN_RETURN_ON_EXCEPTION(isolate, code,
+                             Object::ToString(isolate, code_obj));
   DisplayNamesInternal* internal = display_names->internal()->raw();
   Maybe<icu::UnicodeString> maybe_result =
       internal->of(isolate, code->ToCString().get());
-  MAYBE_RETURN(maybe_result, Handle<Object>());
+  MAYBE_RETURN(maybe_result, DirectHandle<Object>());
   icu::UnicodeString result = maybe_result.FromJust();
   if (result.isBogus()) {
     return isolate->factory()->undefined_value();
@@ -635,34 +635,35 @@ const std::set<std::string>& JSDisplayNames::GetAvailableLocales() {
   return available_locales.Pointer()->Get();
 }
 
-Handle<String> JSDisplayNames::StyleAsString() const {
+Handle<String> JSDisplayNames::StyleAsString(Isolate* isolate) const {
   switch (style()) {
     case Style::kLong:
-      return GetReadOnlyRoots().long_string_handle();
+      return isolate->factory()->long_string();
     case Style::kShort:
-      return GetReadOnlyRoots().short_string_handle();
+      return isolate->factory()->short_string();
     case Style::kNarrow:
-      return GetReadOnlyRoots().narrow_string_handle();
+      return isolate->factory()->narrow_string();
   }
   UNREACHABLE();
 }
 
-Handle<String> JSDisplayNames::FallbackAsString() const {
+Handle<String> JSDisplayNames::FallbackAsString(Isolate* isolate) const {
   switch (fallback()) {
     case Fallback::kCode:
-      return GetReadOnlyRoots().code_string_handle();
+      return isolate->factory()->code_string();
     case Fallback::kNone:
-      return GetReadOnlyRoots().none_string_handle();
+      return isolate->factory()->none_string();
   }
   UNREACHABLE();
 }
 
-Handle<String> JSDisplayNames::LanguageDisplayAsString() const {
+DirectHandle<String> JSDisplayNames::LanguageDisplayAsString(
+    Isolate* isolate) const {
   switch (language_display()) {
     case LanguageDisplay::kDialect:
-      return GetReadOnlyRoots().dialect_string_handle();
+      return isolate->factory()->dialect_string();
     case LanguageDisplay::kStandard:
-      return GetReadOnlyRoots().standard_string_handle();
+      return isolate->factory()->standard_string();
   }
   UNREACHABLE();
 }
