@@ -8,26 +8,22 @@ function blah() {
     var a = 10;
     console.log(a);
   } catch (e) {}
-
   label: for (var i = 0; i < 100; i++) {
     var b = 0;
-
     while (b < 10) {
       console.log(b);
       b += 2;
       continue label;
     }
   }
+  return 1;
 }
-
 blah();
 blah();
-
 (function () {
   1;
   1;
 })();
-
 if (true) {
   2;
   2;
@@ -35,15 +31,46 @@ if (true) {
   3;
   3;
 }
-
 let a = 0;
-
 switch (a) {
   case 1:
     1;
 }
-
 with (Math) {
   cos(PI);
 }
 let module = new WebAssembly.Module(builder.toBuffer());
+const complex1 = [1, 2, 3];
+const complex2 = boom();
+let complex3 = function () {
+  let complex4 = [1, 2, 3];
+  return 2;
+}();
+if (true) {
+  let complex5 = new Map();
+}
+async function foo(a) {
+  let val = await a;
+}
+try {
+  1;
+} catch (e) {
+  2;
+}
+call(() => {
+  1;
+  2;
+});
+function foo() {
+  let a = 0;
+  let b = 1;
+  boom();
+  return {
+    a: a,
+    b: b
+  };
+}
+let {
+  a: x,
+  b: y
+} = foo();
