@@ -9,6 +9,8 @@ const { getValidatedPath } = require('internal/fs/utils');
 const fsBindings = internalBinding('fs');
 const { internal: internalConstants } = internalBinding('constants');
 
+const experimentalAddonModules = getOptionValue('--experimental-addon-modules');
+
 const extensionFormatMap = {
   '__proto__': null,
   '.cjs': 'commonjs',
@@ -18,10 +20,11 @@ const extensionFormatMap = {
   '.wasm': 'wasm',
 };
 
-if (getOptionValue('--experimental-addon-modules')) {
+if (experimentalAddonModules) {
   extensionFormatMap['.node'] = 'addon';
 }
-if (getOptionValue('--experimental-strip-types')) {
+
+if (getOptionValue('--strip-types')) {
   extensionFormatMap['.ts'] = 'module-typescript';
   extensionFormatMap['.mts'] = 'module-typescript';
   extensionFormatMap['.cts'] = 'commonjs-typescript';
@@ -44,10 +47,11 @@ function mimeToFormat(mime) {
 }
 
 /**
- * For extensionless files in a `module` package scope, or a default `module` scope enabled by the
- * `--experimental-default-type` flag, we check the file contents to disambiguate between ES module JavaScript and Wasm.
+ * For extensionless files in a `module` package scope, we check the file contents to disambiguate between ES module
+ * JavaScript and Wasm.
  * We do this by taking advantage of the fact that all Wasm files start with the header `0x00 0x61 0x73 0x6d` (`_asm`).
  * @param {URL} url
+ * @returns {'wasm'|'module'}
  */
 function getFormatOfExtensionlessFile(url) {
   const path = getValidatedPath(url);
