@@ -7,6 +7,8 @@ const {
   FunctionPrototypeCall,
   JSONStringify,
   ObjectSetPrototypeOf,
+  Promise,
+  PromisePrototypeThen,
   RegExpPrototypeSymbolReplace,
   encodeURIComponent,
   hardenRegExp,
@@ -25,17 +27,20 @@ const {
   ERR_REQUIRE_ASYNC_MODULE,
   ERR_REQUIRE_CYCLE_MODULE,
   ERR_REQUIRE_ESM,
-  ERR_NETWORK_IMPORT_DISALLOWED,
   ERR_UNKNOWN_MODULE_FORMAT,
 } = require('internal/errors').codes;
 const { getOptionValue } = require('internal/options');
-const { isURL, pathToFileURL, URLParse } = require('internal/url');
+const { isURL, pathToFileURL } = require('internal/url');
 const { kEmptyObject } = require('internal/util');
 const {
   compileSourceTextModule,
   getDefaultConditions,
   shouldSpawnLoaderHookWorker,
+  requestTypes: { kImportInRequiredESM, kRequireInImportedCJS, kImportInImportedESM },
 } = require('internal/modules/esm/utils');
+/**
+ * @typedef {import('./utils.js').ModuleRequestType} ModuleRequestType
+ */
 const { kImplicitTypeAttribute } = require('internal/modules/esm/assert');
 const {
   ModuleWrap,
@@ -111,15 +116,16 @@ function getTranslators() {
  * async linking.
  * @param {string} filename Filename of the module being required.
  * @param {string|undefined} parentFilename Filename of the module calling require().
+ * @param {boolean} isForAsyncLoaderHookWorker Whether this is for the async loader hook worker.
  * @returns {string} Error message.
  */
-function getRaceMessage(filename, parentFilename) {
-  let raceMessage = `Cannot require() ES Module ${filename} because it is not yet fully loaded. `;
+function getRaceMessage(filename, parentFilename, isForAsyncLoaderHookWorker) {
+  let raceMessage = `Cannot require() ES Module ${filename} because it is not yet fully loaded.\n`;
   raceMessage += 'This may be caused by a race condition if the module is simultaneously dynamically ';
-  raceMessage += 'import()-ed via Promise.all(). Try await-ing the import() sequentially in a loop instead.';
-  if (parentFilename) {
-    raceMessage += ` (from ${parentFilename})`;
-  }
+  raceMessage += 'import()-ed via Promise.all().\n';
+  raceMessage += 'Try await-ing the import() sequentially in a loop instead.\n';
+  raceMessage += ` (From ${parentFilename ? `${parentFilename} in ` : ' '}`;
+  raceMessage += `${isForAsyncLoaderHookWorker ? 'loader hook worker thread' : 'non-loader-hook thread'})`;
   return raceMessage;
 }
 
@@ -184,6 +190,12 @@ class ModuleLoader {
    */
   allowImportMetaResolve;
 
+  /**
+   * @see {AsyncLoaderHooks.isForAsyncLoaderHookWorker}
+   * Shortcut to this.#asyncLoaderHooks.isForAsyncLoaderHookWorker.
+   */
+  isForAsyncLoaderHookWorker = false;
+
   /**
    * Asynchronous loader hooks to pass requests to.
    *
@@ -223,8 +235,10 @@ class ModuleLoader {
     this.#asyncLoaderHooks = asyncLoaderHooks;
     if (asyncLoaderHooks) {
       this.allowImportMetaResolve = asyncLoaderHooks.allowImportMetaResolve;
+      this.isForAsyncLoaderHookWorker = asyncLoaderHooks.isForAsyncLoaderHookWorker;
     } else {
       this.allowImportMetaResolve = true;
+      this.isForAsyncLoaderHookWorker = false;
     }
   }
 
@@ -249,7 +263,7 @@ class ModuleLoader {
   async executeModuleJob(url, wrap, isEntryPoint = false) {
     const { ModuleJob } = require('internal/modules/esm/module_job');
     const module = await onImport.tracePromise(async () => {
-      const job = new ModuleJob(this, url, undefined, wrap, kEvaluationPhase, false, false);
+      const job = new ModuleJob(this, url, undefined, wrap, kEvaluationPhase, false, false, kImportInImportedESM);
       this.loadCache.set(url, undefined, job);
       const { module } = await job.run(isEntryPoint);
       return module;
@@ -279,62 +293,6 @@ class ModuleLoader {
     return this.executeModuleJob(url, wrap, isEntryPoint);
   }
 
-  /**
-   * Get a (possibly not yet fully linked) module job from the cache, or create one and return its Promise.
-   * @param {string} specifier The module request of the module to be resolved. Typically, what's
-   *   requested by `import '<specifier>'` or `import('<specifier>')`.
-   * @param {string} [parentURL] The URL of the module where the module request is initiated.
-   *   It's undefined if it's from the root module.
-   * @param {ImportAttributes} importAttributes Attributes from the import statement or expression.
-   * @param {number} phase Import phase.
-   * @returns {Promise<ModuleJobBase>}
-   */
-  async getModuleJobForImport(specifier, parentURL, importAttributes, phase) {
-    const resolveResult = await this.resolve(specifier, parentURL, importAttributes);
-    return this.#getJobFromResolveResult(resolveResult, parentURL, importAttributes, phase, false);
-  }
-
-  /**
-   * Similar to {@link getModuleJobForImport} but it's used for `require()` resolved by the ESM loader
-   * in imported CJS modules. This runs synchronously and when it returns, the module job's module
-   * requests are all linked.
-   * @param {string} specifier See {@link getModuleJobForImport}
-   * @param {string} [parentURL] See {@link getModuleJobForImport}
-   * @param {ImportAttributes} importAttributes See {@link getModuleJobForImport}
-   * @param {number} phase Import phase.
-   * @returns {Promise<ModuleJobBase>}
-   */
-  getModuleJobForRequireInImportedCJS(specifier, parentURL, importAttributes, phase) {
-    const resolveResult = this.resolveSync(specifier, parentURL, importAttributes);
-    return this.#getJobFromResolveResult(resolveResult, parentURL, importAttributes, phase, true);
-  }
-
-  /**
-   * Given a resolved module request, obtain a ModuleJobBase from it - if it's already cached,
-   * return the cached ModuleJobBase. Otherwise, load its source and translate it into a ModuleWrap first.
-   * @param {{ format: string, url: string }} resolveResult Resolved module request.
-   * @param {string} [parentURL] See {@link getModuleJobForImport}
-   * @param {ImportAttributes} importAttributes See {@link getModuleJobForImport}
-   * @param {number} phase Import phase.
-   * @param {boolean} isForRequireInImportedCJS Whether this is done for require() in imported CJS.
-   * @returns {ModuleJobBase}
-   */
-  #getJobFromResolveResult(resolveResult, parentURL, importAttributes, phase,
-                           isForRequireInImportedCJS = false) {
-    const { url, format } = resolveResult;
-    const resolvedImportAttributes = resolveResult.importAttributes ?? importAttributes;
-    let job = this.loadCache.get(url, resolvedImportAttributes.type);
-
-    if (job === undefined) {
-      job = this.#createModuleJob(url, resolvedImportAttributes, phase, parentURL, format,
-                                  isForRequireInImportedCJS);
-    } else {
-      job.ensurePhase(phase);
-    }
-
-    return job;
-  }
-
   /**
    * This constructs (creates, instantiates and evaluates) a module graph that
    * is require()'d.
@@ -347,6 +305,10 @@ class ModuleLoader {
    */
   importSyncForRequire(mod, filename, source, isMain, parent) {
     const url = pathToFileURL(filename).href;
+    if (!getOptionValue('--experimental-require-module')) {
+      throw new ERR_REQUIRE_ESM(url, true);
+    }
+
     let job = this.loadCache.get(url, kImplicitTypeAttribute);
     // This module job is already created:
     // 1. If it was loaded by `require()` before, at this point the instantiation
@@ -364,9 +326,9 @@ class ModuleLoader {
     if (job !== undefined) {
       mod[kRequiredModuleSymbol] = job.module;
       const parentFilename = urlToFilename(parent?.filename);
-      // TODO(node:55782): this race may stop to happen when the ESM resolution and loading become synchronous.
+      // This race should only be possible on the loader hook thread. See https://github.com/nodejs/node/issues/59666
       if (!job.module) {
-        assert.fail(getRaceMessage(filename, parentFilename));
+        assert.fail(getRaceMessage(filename, parentFilename), this.isForAsyncLoaderHookWorker);
       }
       const status = job.module.getStatus();
       debug('Module status', job, status);
@@ -419,92 +381,67 @@ class ModuleLoader {
     const inspectBrk = (isMain && getOptionValue('--inspect-brk'));
 
     const { ModuleJobSync } = require('internal/modules/esm/module_job');
-    job = new ModuleJobSync(this, url, kEmptyObject, wrap, kEvaluationPhase, isMain, inspectBrk);
+    job = new ModuleJobSync(this, url, kEmptyObject, wrap, kEvaluationPhase, isMain, inspectBrk,
+                            kImportInRequiredESM);
     this.loadCache.set(url, kImplicitTypeAttribute, job);
     mod[kRequiredModuleSymbol] = job.module;
     return { wrap: job.module, namespace: job.runSync(parent).namespace };
   }
 
   /**
-   * Resolve individual module requests and create or get the cached ModuleWraps for
-   * each of them. This is only used to create a module graph being require()'d.
-   * @param {string} specifier Specifier of the the imported module.
-   * @param {string} parentURL Where the import comes from.
-   * @param {object} importAttributes import attributes from the import statement.
-   * @param {number} phase The import phase.
-   * @returns {ModuleJobBase}
+   * Check invariants on a cached module job when require()'d from ESM.
+   * @param {string} specifier The first parameter of require().
+   * @param {string} url URL of the module being required.
+   * @param {string|undefined} parentURL URL of the module calling require().
+   * @param {ModuleJobBase} job The cached module job.
    */
-  getModuleJobForRequire(specifier, parentURL, importAttributes, phase) {
-    const parsed = URLParse(specifier);
-    if (parsed != null) {
-      const protocol = parsed.protocol;
-      if (protocol === 'https:' || protocol === 'http:') {
-        throw new ERR_NETWORK_IMPORT_DISALLOWED(specifier, parentURL,
-                                                'ES modules cannot be loaded by require() from the network');
+  #checkCachedJobForRequireESM(specifier, url, parentURL, job) {
+    // This race should only be possible on the loader hook thread. See https://github.com/nodejs/node/issues/59666
+    if (!job.module) {
+      assert.fail(getRaceMessage(url, parentURL, this.isForAsyncLoaderHookWorker));
+    }
+    // This module is being evaluated, which means it's imported in a previous link
+    // in a cycle.
+    if (job.module.getStatus() === kEvaluating) {
+      const parentFilename = urlToFilename(parentURL);
+      let message = `Cannot import Module ${specifier} in a cycle.`;
+      if (parentFilename) {
+        message += ` (from ${parentFilename})`;
       }
-      assert(protocol === 'file:' || protocol === 'node:' || protocol === 'data:');
+      throw new ERR_REQUIRE_CYCLE_MODULE(message);
     }
 
-    // TODO(joyeecheung): consolidate cache behavior and use resolveSync() and
-    // loadSync() here.
-    const resolveResult = this.#cachedResolveSync(specifier, parentURL, importAttributes);
-    const { url, format } = resolveResult;
-    if (!getOptionValue('--experimental-require-module')) {
-      throw new ERR_REQUIRE_ESM(url, true);
-    }
-    const resolvedImportAttributes = resolveResult.importAttributes ?? importAttributes;
-    let job = this.loadCache.get(url, resolvedImportAttributes.type);
-    if (job !== undefined) {
-      // TODO(node:55782): this race may stop happening when the ESM resolution and loading become synchronous.
-      if (!job.module) {
-        assert.fail(getRaceMessage(url, parentURL));
-      }
-      // This module is being evaluated, which means it's imported in a previous link
-      // in a cycle.
-      if (job.module.getStatus() === kEvaluating) {
-        const parentFilename = urlToFilename(parentURL);
-        let message = `Cannot import Module ${specifier} in a cycle.`;
-        if (parentFilename) {
-          message += ` (from ${parentFilename})`;
-        }
-        throw new ERR_REQUIRE_CYCLE_MODULE(message);
-      }
-      job.ensurePhase(phase);
-      // Otherwise the module could be imported before but the evaluation may be already
-      // completed (e.g. the require call is lazy) so it's okay. We will return the
-      // module now and check asynchronicity of the entire graph later, after the
-      // graph is instantiated.
-      return job;
-    }
+    // Otherwise the module could be imported before but the evaluation may be already
+    // completed (e.g. the require call is lazy) so it's okay. We will return the
+    // job and check asynchronicity of the entire graph later, after the
+    // graph is instantiated.
+  }
 
-    const loadResult = this.#loadSync(url, { format, importAttributes });
-
-    const formatFromLoad = loadResult.format;
+  /**
+   * Load a module and translate it into a ModuleWrap for require(esm).
+   * This is run synchronously, and the translator always return a ModuleWrap synchronously.
+   * @param {string} url URL of the module to be translated.
+   * @param {object} loadContext See {@link load}
+   * @param {string|undefined} parentURL URL of the parent module. Undefined if it's the entry point.
+   * @param {ModuleRequest} request Module request.
+   * @returns {ModuleWrap}
+   */
+  loadAndTranslateForImportInRequiredESM(url, loadContext, parentURL, request) {
+    const loadResult = this.#loadSync(url, loadContext);
     // Use the synchronous commonjs translator which can deal with cycles.
+    const formatFromLoad = loadResult.format;
     const translatorKey = (formatFromLoad === 'commonjs' || formatFromLoad === 'commonjs-typescript') ?
       'commonjs-sync' : formatFromLoad;
-
-    if (translatorKey === 'wasm') {
-      assert.fail('WASM is currently unsupported by require(esm)');
-    }
-
-    const { source } = loadResult;
-    const isMain = (parentURL === undefined);
-    const translateContext = { format: formatFromLoad, source, translatorKey, __proto__: null };
+    const translateContext = { ...loadResult, translatorKey, __proto__: null };
     const wrap = this.#translate(url, translateContext, parentURL);
     assert(wrap instanceof ModuleWrap, `Translator used for require(${url}) should not be async`);
 
-    if (process.env.WATCH_REPORT_DEPENDENCIES && process.send) {
-      process.send({ 'watch:import': [url] });
-    }
-
     const cjsModule = wrap[imported_cjs_symbol];
     if (cjsModule) {
-      assert(translatorKey === 'commonjs-sync');
       // Check if the ESM initiating import CJS is being required by the same CJS module.
       if (cjsModule?.[kIsExecuting]) {
         const parentFilename = urlToFilename(parentURL);
-        let message = `Cannot import CommonJS Module ${specifier} in a cycle.`;
+        let message = `Cannot import CommonJS Module ${request.specifier} in a cycle.`;
         if (parentFilename) {
           message += ` (from ${parentFilename})`;
         }
@@ -512,12 +449,7 @@ class ModuleLoader {
       }
     }
 
-    const inspectBrk = (isMain && getOptionValue('--inspect-brk'));
-    const { ModuleJobSync } = require('internal/modules/esm/module_job');
-    job = new ModuleJobSync(this, url, importAttributes, wrap, phase, isMain, inspectBrk);
-
-    this.loadCache.set(url, importAttributes.type, job);
-    return job;
+    return wrap;
   }
 
   /**
@@ -540,6 +472,9 @@ class ModuleLoader {
 
     const result = FunctionPrototypeCall(translator, this, url, translateContext, parentURL);
     assert(result instanceof ModuleWrap, `The ${format} module returned is not a ModuleWrap`);
+    if (format === 'commonjs' || format === 'commonjs-sync' || format === 'require-commonjs') {
+      result.isCommonJS = true;
+    }
     return result;
   }
 
@@ -594,54 +529,72 @@ class ModuleLoader {
       return this.#translate(url, translateContext, parentURL);
     };
     if (isPromise(maybePromise)) {
-      return maybePromise.then(afterLoad);
+      return PromisePrototypeThen(maybePromise, afterLoad);
     }
     return afterLoad(maybePromise);
   }
 
   /**
-   * Load a module and translate it into a ModuleWrap, and create a ModuleJob from it.
-   * This runs synchronously. If isForRequireInImportedCJS is true, the module should be linked
-   * by the time this returns. Otherwise it may still have pending module requests.
-   * @param {string} url The URL that was resolved for this module.
-   * @param {ImportAttributes} importAttributes See {@link getModuleJobForImport}
-   * @param {number} phase Import phase.
-   * @param {string} [parentURL] See {@link getModuleJobForImport}
-   * @param {string} [format] The format hint possibly returned by the `resolve` hook
-   * @param {boolean} isForRequireInImportedCJS Whether this module job is created for require()
-   *   in imported CJS.
+   * Given a resolved module request, obtain a ModuleJobBase from it - if it's already cached,
+   * return the cached ModuleJobBase. Otherwise, load its source and translate it into a ModuleWrap first.
+   * This runs synchronously. On any thread that is not an async loader hook worker thread,
+   * the module should be linked by the time this returns. Otherwise it may still have
+   * pending module requests.
+   * @param {string} parentURL See {@link getOrCreateModuleJob}
+   * @param {{format: string, url: string}} resolveResult
+   * @param {ModuleRequest} request Module request.
+   * @param {ModuleRequestType} requestType Type of the module request.
    * @returns {ModuleJobBase} The (possibly pending) module job
    */
-  #createModuleJob(url, importAttributes, phase, parentURL, format, isForRequireInImportedCJS) {
-    const context = { format, importAttributes };
+  #getOrCreateModuleJobAfterResolve(parentURL, resolveResult, request, requestType) {
+    const { url, format } = resolveResult;
+    // TODO(joyeecheung): update the module requests to use importAttributes as property names.
+    const importAttributes = resolveResult.importAttributes ?? request.attributes;
+    let job = this.loadCache.get(url, importAttributes.type);
+
+    if (job !== undefined) {
+      if (requestType === kImportInRequiredESM) {
+        this.#checkCachedJobForRequireESM(request.specifier, url, parentURL, job);
+      }
+      job.ensurePhase(request.phase, requestType);
+      return job;
+    }
+
+    const context = { format, importAttributes, __proto__: null };
 
-    const isMain = parentURL === undefined;
     let moduleOrModulePromise;
-    if (isForRequireInImportedCJS) {
+    if (requestType === kRequireInImportedCJS) {
       moduleOrModulePromise = this.loadAndTranslateForRequireInImportedCJS(url, context, parentURL);
+    } else if (requestType === kImportInRequiredESM) {
+      moduleOrModulePromise = this.loadAndTranslateForImportInRequiredESM(url, context, parentURL, request);
     } else {
       moduleOrModulePromise = this.loadAndTranslate(url, context, parentURL);
     }
 
-    const inspectBrk = (
-      isMain &&
-      getOptionValue('--inspect-brk')
-    );
-
-    if (process.env.WATCH_REPORT_DEPENDENCIES && process.send) {
-      process.send({ 'watch:import': [url] });
+    if (requestType === kImportInRequiredESM || requestType === kRequireInImportedCJS ||
+        !this.isForAsyncLoaderHookWorker) {
+      assert(moduleOrModulePromise instanceof ModuleWrap, `Expected ModuleWrap for loading ${url}`);
     }
 
-    const { ModuleJob } = require('internal/modules/esm/module_job');
-    const job = new ModuleJob(
+    if (process.env.WATCH_REPORT_DEPENDENCIES && process.send) {
+      const type = requestType === kRequireInImportedCJS ? 'require' : 'import';
+      process.send({ [`watch:${type}`]: [url] });
+    }
+
+    const { ModuleJob, ModuleJobSync } = require('internal/modules/esm/module_job');
+    // TODO(joyeecheung): use ModuleJobSync for kRequireInImportedCJS too.
+    const ModuleJobCtor = (requestType === kImportInRequiredESM ? ModuleJobSync : ModuleJob);
+    const isMain = (parentURL === undefined);
+    const inspectBrk = (isMain && getOptionValue('--inspect-brk'));
+    job = new ModuleJobCtor(
       this,
       url,
       importAttributes,
       moduleOrModulePromise,
-      phase,
+      request.phase,
       isMain,
       inspectBrk,
-      isForRequireInImportedCJS,
+      requestType,
     );
 
     this.loadCache.set(url, importAttributes.type, job);
@@ -649,6 +602,34 @@ class ModuleLoader {
     return job;
   }
 
+  /**
+   * Get a (possibly not yet fully linked) module job from the cache, or create one and return its Promise.
+   * @param {string} [parentURL] The URL of the module where the module request is initiated.
+   *   It's undefined if it's from the root module.
+   * @param {ModuleRequest} request Module request.
+   * @param {ModuleRequestType} requestType Type of the module request.
+   * @returns {Promise<ModuleJobBase>|ModuleJobBase}
+   */
+  getOrCreateModuleJob(parentURL, request, requestType) {
+    let maybePromise;
+    if (requestType === kRequireInImportedCJS || requestType === kImportInRequiredESM) {
+      // In these two cases, resolution must be synchronous.
+      maybePromise = this.resolveSync(parentURL, request);
+      assert(!isPromise(maybePromise));
+    } else {
+      maybePromise = this.#resolve(parentURL, request);
+    }
+
+    const afterResolve = (resolveResult) => {
+      return this.#getOrCreateModuleJobAfterResolve(parentURL, resolveResult, request, requestType);
+    };
+
+    if (isPromise(maybePromise)) {
+      return PromisePrototypeThen(maybePromise, afterResolve);
+    }
+    return afterResolve(maybePromise);
+  }
+
   /**
    * This method is usually called indirectly as part of the loading processes.
    * Use directly with caution.
@@ -662,8 +643,16 @@ class ModuleLoader {
    */
   async import(specifier, parentURL, importAttributes, phase = kEvaluationPhase, isEntryPoint = false) {
     return onImport.tracePromise(async () => {
-      const moduleJob = await this.getModuleJobForImport(specifier, parentURL, importAttributes,
-                                                         phase);
+      const request = { specifier, phase, attributes: importAttributes, __proto__: null };
+      let moduleJob;
+      try {
+        moduleJob = await this.getOrCreateModuleJob(parentURL, request);
+      } catch (e) {
+        if (e?.code === 'ERR_ASYNC_LOADER_REQUEST_NEVER_SETTLED') {
+          return new Promise(() => {});
+        }
+        throw e;
+      }
       if (phase === kSourcePhase) {
         const module = await moduleJob.modulePromise;
         return module.getModuleSourceObject();
@@ -695,28 +684,20 @@ class ModuleLoader {
   /**
    * Resolve a module request to a URL identifying the location of the module. Handles customization hooks,
    * if any.
-   * @param {string|URL} specifier The module request of the module to be resolved. Typically, what's
-   *   requested by `import specifier`, `import(specifier)` or `import.meta.resolve(specifier)`.
    * @param {string} [parentURL] The URL of the module where the module request is initiated.
    *   It's undefined if it's from the root module.
-   * @param {ImportAttributes} importAttributes Attributes from the import statement or expression.
-   * @returns {Promise<{format: string, url: string}>}
+   * @param {ModuleRequest} request Module request.
+   * @returns {Promise<{format: string, url: string}>|{format: string, url: string}}
    */
-  resolve(specifier, parentURL, importAttributes) {
-    specifier = `${specifier}`;
-    if (syncResolveHooks.length) {
-      // Has module.registerHooks() hooks, use the synchronous variant that can handle both hooks.
-      return this.resolveSync(specifier, parentURL, importAttributes);
-    }
-    if (this.#asyncLoaderHooks) {  // Only has module.register hooks.
+  #resolve(parentURL, request) {
+    if (this.isForAsyncLoaderHookWorker) {
+      const specifier = `${request.specifier}`;
+      const importAttributes = request.attributes ?? kEmptyObject;
+      // TODO(joyeecheung): invoke the synchronous hooks in the default step on loader thread.
       return this.#asyncLoaderHooks.resolve(specifier, parentURL, importAttributes);
     }
-    return this.#cachedDefaultResolve(specifier, {
-      __proto__: null,
-      conditions: this.#defaultConditions,
-      parentURL,
-      importAttributes,
-    });
+
+    return this.resolveSync(parentURL, request);
   }
 
   /**
@@ -739,25 +720,6 @@ class ModuleLoader {
     return result;
   }
 
-  /**
-   * Either return a cached resolution, or perform the synchronous resolution, and
-   * cache the result.
-   * @param {string} specifier See {@link resolve}.
-   * @param {string} [parentURL] See {@link resolve}.
-   * @param {ImportAttributes} importAttributes See {@link resolve}.
-   * @returns {{ format: string, url: string }}
-   */
-  #cachedResolveSync(specifier, parentURL, importAttributes) {
-    const requestKey = this.#resolveCache.serializeKey(specifier, importAttributes);
-    const cachedResult = this.#resolveCache.get(requestKey, parentURL);
-    if (cachedResult != null) {
-      return cachedResult;
-    }
-    const result = this.resolveSync(specifier, parentURL, importAttributes);
-    this.#resolveCache.set(requestKey, parentURL, result);
-    return result;
-  }
-
   /**
    * This is the default resolve step for module.registerHooks(), which incorporates asynchronous hooks
    * from module.register() which are run in a blocking fashion for it to be synchronous.
@@ -767,7 +729,7 @@ class ModuleLoader {
    * @returns {{ format: string, url: string }}
    */
   #resolveAndMaybeBlockOnLoaderThread(specifier, context) {
-    if (this.#asyncLoaderHooks) {
+    if (this.#asyncLoaderHooks?.resolveSync) {
       return this.#asyncLoaderHooks.resolveSync(specifier, context.parentURL, context.importAttributes);
     }
     return this.#cachedDefaultResolve(specifier, context);
@@ -779,46 +741,43 @@ class ModuleLoader {
    * from the loader thread for this to be synchronous.
    * This is here to support `import.meta.resolve()`, `require()` in imported CJS, and
    * `module.registerHooks()` hooks.
-   *
-   * TODO(joyeecheung): consolidate the cache behavior and use this in require(esm).
-   * @param {string|URL} specifier See {@link resolve}.
    * @param {string} [parentURL] See {@link resolve}.
-   * @param {ImportAttributes} [importAttributes] See {@link resolve}.
+   * @param {ModuleRequest} request See {@link resolve}.
    * @returns {{ format: string, url: string }}
    */
-  resolveSync(specifier, parentURL, importAttributes = { __proto__: null }) {
-    specifier = `${specifier}`;
+  resolveSync(parentURL, request) {
+    const specifier = `${request.specifier}`;
+    const importAttributes = request.attributes ?? kEmptyObject;
+
     if (syncResolveHooks.length) {
       // Has module.registerHooks() hooks, chain the asynchronous hooks in the default step.
       return resolveWithSyncHooks(specifier, parentURL, importAttributes, this.#defaultConditions,
                                   this.#resolveAndMaybeBlockOnLoaderThread.bind(this));
     }
-    return this.#resolveAndMaybeBlockOnLoaderThread(specifier, {
-      __proto__: null,
+    const context = {
+      ...request,
       conditions: this.#defaultConditions,
       parentURL,
       importAttributes,
-    });
+      __proto__: null,
+    };
+    return this.#resolveAndMaybeBlockOnLoaderThread(specifier, context);
   }
 
   /**
    * Provide source that is understood by one of Node's translators. Handles customization hooks,
    * if any.
+   * @typedef { {format: ModuleFormat, source: ModuleSource }} LoadResult
    * @param {string} url The URL of the module to be loaded.
    * @param {object} context Metadata about the module
-   * @returns {Promise<{ format: ModuleFormat, source: ModuleSource }> | { format: ModuleFormat, source: ModuleSource }}
+   * @returns {Promise<LoadResult> | LoadResult}}
    */
   load(url, context) {
-    if (syncLoadHooks.length) {
-      // Has module.registerHooks() hooks, use the synchronous variant that can handle both hooks.
-      return this.#loadSync(url, context);
-    }
-    if (this.#asyncLoaderHooks) {
+    if (this.isForAsyncLoaderHookWorker) {
+      // TODO(joyeecheung): invoke the synchronous hooks in the default step on loader thread.
       return this.#asyncLoaderHooks.load(url, context);
     }
-
-    defaultLoadSync ??= require('internal/modules/esm/load').defaultLoadSync;
-    return defaultLoadSync(url, context);
+    return this.#loadSync(url, context);
   }
 
   /**
@@ -829,7 +788,7 @@ class ModuleLoader {
    * @returns {{ format: ModuleFormat, source: ModuleSource }}
    */
   #loadAndMaybeBlockOnLoaderThread(url, context) {
-    if (this.#asyncLoaderHooks) {
+    if (this.#asyncLoaderHooks?.loadSync) {
       return this.#asyncLoaderHooks.loadSync(url, context);
     }
     defaultLoadSync ??= require('internal/modules/esm/load').defaultLoadSync;
@@ -841,8 +800,6 @@ class ModuleLoader {
    * from module.register(), this blocks on the loader thread for it to return synchronously.
    *
    * This is here to support `require()` in imported CJS and `module.registerHooks()` hooks.
-   *
-   * TODO(joyeecheung): consolidate the cache behavior and use this in require(esm).
    * @param {string} url See {@link load}
    * @param {object} [context] See {@link load}
    * @returns {{ format: ModuleFormat, source: ModuleSource }}
