@@ -5,6 +5,7 @@ const {
   ObjectFreeze,
   SafeSet,
   SafeWeakMap,
+  Symbol,
 } = primordials;
 
 const {
@@ -332,6 +333,16 @@ function compileSourceTextModule(url, source, cascadedLoader, context = kEmptyOb
   return wrap;
 }
 
+
+const kImportInImportedESM = Symbol('kImportInImportedESM');
+const kImportInRequiredESM = Symbol('kImportInRequiredESM');
+const kRequireInImportedCJS = Symbol('kRequireInImportedCJS');
+
+/**
+ * @typedef {kImportInImportedESM | kImportInRequiredESM | kRequireInImportedCJS} ModuleRequestType
+ */
+const requestTypes = { kImportInImportedESM, kImportInRequiredESM, kRequireInImportedCJS };
+
 module.exports = {
   registerModule,
   initializeESM,
@@ -339,4 +350,5 @@ module.exports = {
   getConditionsSet,
   shouldSpawnLoaderHookWorker,
   compileSourceTextModule,
+  requestTypes,
 };
