@@ -28,6 +28,8 @@ const {
 } = require('internal/histogram');
 
 const monitorEventLoopDelay = require('internal/perf/event_loop_delay');
+const { eventLoopUtilization } = require('internal/perf/event_loop_utilization');
+const timerify = require('internal/perf/timerify');
 
 module.exports = {
   Performance,
@@ -38,6 +40,8 @@ module.exports = {
   PerformanceObserverEntryList,
   PerformanceResourceTiming,
   monitorEventLoopDelay,
+  eventLoopUtilization,
+  timerify,
   createHistogram,
   performance,
 };
