@@ -103,67 +103,69 @@ SerializerContext::SerializerContext(Environment* env, Local<Object> wrap)
 
 void SerializerContext::ThrowDataCloneError(Local<String> message) {
   Local<Value> args[1] = { message };
-  Local<Value> get_data_clone_error =
-      object()->Get(env()->context(),
-                    env()->get_data_clone_error_string())
-                      .ToLocalChecked();
+  Local<Value> get_data_clone_error;
+  if (!object()
+           ->Get(env()->context(), env()->get_data_clone_error_string())
+           .ToLocal(&get_data_clone_error)) {
+    // A superseding error will have been thrown by v8.
+    return;
+  }
 
   CHECK(get_data_clone_error->IsFunction());
-  MaybeLocal<Value> error =
-      get_data_clone_error.As<Function>()->Call(env()->context(),
-                                                object(),
-                                                arraysize(args),
-                                                args);
-
-  if (error.IsEmpty()) return;
-
-  env()->isolate()->ThrowException(error.ToLocalChecked());
+  Local<Value> error;
+  if (get_data_clone_error.As<Function>()
+          ->Call(env()->context(), object(), arraysize(args), args)
+          .ToLocal(&error)) {
+    env()->isolate()->ThrowException(error);
+  }
 }
 
 Maybe<uint32_t> SerializerContext::GetSharedArrayBufferId(
     Isolate* isolate, Local<SharedArrayBuffer> shared_array_buffer) {
   Local<Value> args[1] = { shared_array_buffer };
-  Local<Value> get_shared_array_buffer_id =
-      object()->Get(env()->context(),
-                    env()->get_shared_array_buffer_id_string())
-                      .ToLocalChecked();
+  Local<Value> get_shared_array_buffer_id;
+  if (!object()
+           ->Get(env()->context(), env()->get_shared_array_buffer_id_string())
+           .ToLocal(&get_shared_array_buffer_id)) {
+    return Nothing<uint32_t>();
+  }
 
   if (!get_shared_array_buffer_id->IsFunction()) {
     return ValueSerializer::Delegate::GetSharedArrayBufferId(
         isolate, shared_array_buffer);
   }
 
-  MaybeLocal<Value> id =
-      get_shared_array_buffer_id.As<Function>()->Call(env()->context(),
-                                                      object(),
-                                                      arraysize(args),
-                                                      args);
+  Local<Value> id;
+  if (!get_shared_array_buffer_id.As<Function>()
+           ->Call(env()->context(), object(), arraysize(args), args)
+           .ToLocal(&id)) {
+    return Nothing<uint32_t>();
+  }
 
-  if (id.IsEmpty()) return Nothing<uint32_t>();
-
-  return id.ToLocalChecked()->Uint32Value(env()->context());
+  return id->Uint32Value(env()->context());
 }
 
 Maybe<bool> SerializerContext::WriteHostObject(Isolate* isolate,
                                                Local<Object> input) {
-  MaybeLocal<Value> ret;
   Local<Value> args[1] = { input };
 
-  Local<Value> write_host_object =
-      object()->Get(env()->context(),
-                    env()->write_host_object_string()).ToLocalChecked();
+  Local<Value> write_host_object;
+  if (!object()
+           ->Get(env()->context(), env()->write_host_object_string())
+           .ToLocal(&write_host_object)) {
+    return Nothing<bool>();
+  }
 
   if (!write_host_object->IsFunction()) {
     return ValueSerializer::Delegate::WriteHostObject(isolate, input);
   }
 
-  ret = write_host_object.As<Function>()->Call(env()->context(),
-                                               object(),
-                                               arraysize(args),
-                                               args);
-
-  if (ret.IsEmpty())
+  Local<Value> ret;
+  if (!write_host_object.As<Function>()
+           ->Call(env()->context(), object(), arraysize(args), args)
+           .ToLocal(&ret)) {
     return Nothing<bool>();
+  }
 
   return Just(true);
 }
@@ -209,12 +211,10 @@ void SerializerContext::ReleaseBuffer(const FunctionCallbackInfo<Value>& args) {
   // Note: Both ValueSerializer and this Buffer::New() variant use malloc()
   // as the underlying allocator.
   std::pair<uint8_t*, size_t> ret = ctx->serializer_.Release();
-  auto buf = Buffer::New(ctx->env(),
-                         reinterpret_cast<char*>(ret.first),
-                         ret.second);
-
-  if (!buf.IsEmpty()) {
-    args.GetReturnValue().Set(buf.ToLocalChecked());
+  Local<Object> buf;
+  if (Buffer::New(ctx->env(), reinterpret_cast<char*>(ret.first), ret.second)
+          .ToLocal(&buf)) {
+    args.GetReturnValue().Set(buf);
   }
 }
 
@@ -300,31 +300,31 @@ DeserializerContext::DeserializerContext(Environment* env,
 }
 
 MaybeLocal<Object> DeserializerContext::ReadHostObject(Isolate* isolate) {
-  Local<Value> read_host_object =
-      object()->Get(env()->context(),
-                    env()->read_host_object_string()).ToLocalChecked();
+  Local<Value> read_host_object;
+  if (!object()
+           ->Get(env()->context(), env()->read_host_object_string())
+           .ToLocal(&read_host_object)) {
+    return {};
+  }
 
   if (!read_host_object->IsFunction()) {
     return ValueDeserializer::Delegate::ReadHostObject(isolate);
   }
 
   Isolate::AllowJavascriptExecutionScope allow_js(isolate);
-  MaybeLocal<Value> ret =
-      read_host_object.As<Function>()->Call(env()->context(),
-                                            object(),
-                                            0,
-                                            nullptr);
-
-  if (ret.IsEmpty())
-    return MaybeLocal<Object>();
-
-  Local<Value> return_value = ret.ToLocalChecked();
-  if (!return_value->IsObject()) {
-    env()->ThrowTypeError("readHostObject must return an object");
-    return MaybeLocal<Object>();
+  Local<Value> ret;
+  if (!read_host_object.As<Function>()
+           ->Call(env()->context(), object(), 0, nullptr)
+           .ToLocal(&ret)) {
+    return {};
   }
 
-  return return_value.As<Object>();
+  if (!ret->IsObject()) {
+    env()->ThrowTypeError("readHostObject must return an object");
+    return {};
+  }
+
+  return ret.As<Object>();
 }
 
 void DeserializerContext::New(const FunctionCallbackInfo<Value>& args) {
@@ -356,9 +356,10 @@ void DeserializerContext::ReadValue(const FunctionCallbackInfo<Value>& args) {
   DeserializerContext* ctx;
   ASSIGN_OR_RETURN_UNWRAP(&ctx, args.This());
 
-  MaybeLocal<Value> ret = ctx->deserializer_.ReadValue(ctx->env()->context());
-
-  if (!ret.IsEmpty()) args.GetReturnValue().Set(ret.ToLocalChecked());
+  Local<Value> ret;
+  if (ctx->deserializer_.ReadValue(ctx->env()->context()).ToLocal(&ret)) {
+    args.GetReturnValue().Set(ret);
+  }
 }
 
 void DeserializerContext::TransferArrayBuffer(
