@@ -1,6 +1,6 @@
 'use strict';
 const common = require('../common');
-const { strictEqual } = require('assert');
+const assert = require('assert');
 
 async function test() {
   {
@@ -9,8 +9,8 @@ async function test() {
       import('../fixtures/empty.js'),
     ]);
 
-    strictEqual(results[0].status, 'rejected');
-    strictEqual(results[1].status, 'fulfilled');
+    assert.strictEqual(results[0].status, 'rejected');
+    assert.strictEqual(results[1].status, 'fulfilled');
   }
 
   {
@@ -19,8 +19,8 @@ async function test() {
       import('../fixtures/empty.js', { with: { type: 'json' } }),
     ]);
 
-    strictEqual(results[0].status, 'fulfilled');
-    strictEqual(results[1].status, 'rejected');
+    assert.strictEqual(results[0].status, 'fulfilled');
+    assert.strictEqual(results[1].status, 'rejected');
   }
 
   {
@@ -29,8 +29,8 @@ async function test() {
       import('../fixtures/empty.json'),
     ]);
 
-    strictEqual(results[0].status, 'fulfilled');
-    strictEqual(results[1].status, 'rejected');
+    assert.strictEqual(results[0].status, 'fulfilled');
+    assert.strictEqual(results[1].status, 'rejected');
   }
 
   {
@@ -39,8 +39,8 @@ async function test() {
       import('../fixtures/empty.json', { with: { type: 'json' } }),
     ]);
 
-    strictEqual(results[0].status, 'rejected');
-    strictEqual(results[1].status, 'fulfilled');
+    assert.strictEqual(results[0].status, 'rejected');
+    assert.strictEqual(results[1].status, 'fulfilled');
   }
 }
 
