@@ -7,6 +7,7 @@ const assert = require('assert');
 const warnFlags = [
   '--allow-addons',
   '--allow-child-process',
+  '--allow-inspector',
   '--allow-wasi',
   '--allow-worker',
 ];
@@ -15,7 +16,7 @@ for (const flag of warnFlags) {
   const { status, stderr } = spawnSync(
     process.execPath,
     [
-      '--experimental-permission', flag, '-e',
+      '--permission', flag, '-e',
       'setTimeout(() => {}, 1)',
     ]
   );
